"""MAJIS operations toolbox"""

from .__version__ import __version__
from .itl import Timeline, read_itl, save_itl
from .opl import read_opl, save_opl

__all__ = [
    'read_itl',
    'save_itl',
    'Timeline',
    'read_opl',
    'save_opl',
    '__version__',
]



"""OPL convert module."""

from pathlib import Path

from .export import save_opl_csv, save_opl_json
from .reader import read_opl_csv, read_opl_json


def json2csv_opl(
    opl_json: str | Path | None,
    overlap: bool = False,
    only: str | list | bool | None = 'MAJIS',
) -> Path:
    """Export OPL from JSON to CSV format."""
    return save_opl_csv(
        Path(opl_json).with_suffix('.csv'),
        *read_opl_json(opl_json, only=only),
        overlap=overlap,
    )


def csv2json_opl(
    opl_csv: str | Path | None,
    overlap: bool = False,
) -> Path:
    """Export OPL from CSV to JSON format."""
    return save_opl_json(
        Path(opl_csv).with_suffix('.json'),
        *read_opl_csv(opl_csv),
        overlap=overlap,
    )



"""Depreciation miscellaneous sub-module."""

from functools import wraps
from warnings import warn


def depreciated(func, *, old, new):
    """Depreciation warning decorator."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """Depreciation wrapper."""
        warn(
            f'`{old}` is depreciated in favor of `{new}`.',
            DeprecationWarning,
            stacklevel=2,
        )
        return func(*args, **kwargs)

    return wrapper



"""OPL generic reader module."""

import csv
from pathlib import Path

from planetary_coverage.events import EventsDict, EventsList, EventWindow

from ..itl.reader import read_itl_json
from ..misc.events import flatten, group


def read_opl(
    fname: str | Path,
    flat: bool = False,
    only: str | list | bool | None = 'MAJIS',
) -> EventsDict | EventsList:
    """Read OPL file.

    Both, CSV (``.csv``) and JSON (``.json``) format are supported.

    The result can be return as a nested dict of events (default) or
    as a flat list of events.

    By default only the MAJIS events will be reported.
    This can be changed by providing an other instrument filter key,
    a list of keys or an explicit ``'ALL'``, False or None value.
    Applicable only to JSON OPL.

    """
    f = Path(fname)

    match f.suffix.lower():
        case '.csv':
            events = read_opl_csv(f)
        case '.json':
            events = read_opl_json(f, only=only)
        case _:
            raise ValueError(f'Invalid OPL file input: {fname}')

    return flatten(events) if flat else group(events)


def read_opl_csv(fname: str | Path) -> list[EventWindow]:
    """Read OPL file in CSV format."""
    with Path(fname).open(encoding='utf-8') as content:
        return [EventWindow(**parse_opl_csv(*params)) for params in csv.reader(content)]


def parse_opl_csv(key, start, end, obs_key, instrument) -> dict:
    """OPL CSV line parser.

    Warning
    -------
    Currently the OPL CSV don't have an explicit header.
    Here we assume that the header is as follow"

        # key, start, end, obs_key, instrument

    The initial ``key`` is composed of the following parameters:

        <instrument>_<observation_type>_<type>

    """
    _, observation_type, _type = key.split('_')

    return {
        'key': key,
        't_start': start,
        't_end': end,
        'OBS_KEY': obs_key,
        'INSTRUMENT': instrument,
        'OBSERVATION_TYPE': observation_type,
        'TYPE': _type,
    }


def read_opl_json(
    fname: str | Path,
    only: str | list | bool | None = 'MAJIS',
) -> list[EventWindow]:
    """Read OPL file in JSON format.

    Note
    ----
    Re-use ITL JSON reader with OPL schema instead.

    By default only the MAJIS events will be reported.
    This can be changed by providing an other instrument filter key,
    a list of keys or an explicit ``'ALL'``, False or None value.

    """
    return read_itl_json(fname, fmt='OPL', only=only)



"""MAJIS OPL submodule."""

from .convert import csv2json_opl, json2csv_opl
from .export import save_opl
from .reader import read_opl

__all__ = [
    'read_opl',
    'save_opl',
    'json2csv_opl',
    'csv2json_opl',
]



"""MAJIS ITL submodule."""

from ..misc import depreciated
from .csv import save_itl_csv
from .export import save_itl
from .reader import read_itl
from .timeline import Timeline, save_itl_xlsm

# Depreciated methods
save_csv = depreciated(save_itl_csv, old='save_csv', new='save_itl')
save_xlsm = depreciated(save_itl_xlsm, old='save_xlsm', new='save_itl')

__all__ = [
    'read_itl',
    'save_itl',
    'Timeline',
    'save_csv',  # Depreciated
    'save_xlsm',  # Depreciated
]



"""OPL generic export module."""

from pathlib import Path

from planetary_coverage.events import EventsDict, EventsList, EventWindow

from ..itl.export import save_itl_json
from ..misc import fmt_datetime
from ..misc.events import flatten_events
from ..misc.export import save_file


def save_opl(
    fout: str | Path | None,
    *events: EventsList | EventsDict,
    overlap: bool = False,
    **kwargs,
) -> Path:
    """Save OPL events to a new OPL file.

    Note
    ----
    By default, OPL blocks must not overlap each other.
    This can be disable with ``overlap=True``.

    """
    ext = Path(fout).suffix

    match ext:
        case '.csv':
            return save_opl_csv(
                fout,
                *events,
                overlap=overlap,
                **kwargs,
            )
        case '.json':
            return save_opl_json(
                fout,
                *events,
                overlap=overlap,
                **kwargs,
            )
        case _:
            raise NotImplementedError


@flatten_events
def save_opl_csv(
    fout: str | Path | None,
    *events: EventsList | EventsDict,
    overlap: bool = False,
    sep: str = ',',
) -> Path:
    """Export OPL file in CSV format."""
    content = [sep.join(fmt_opl_csv(event)) for event in events]

    return save_file(fout, content, suffix='.csv')


def fmt_opl_csv(obs: EventWindow) -> list[str]:
    """OPL CSV line formatter.

    Warning
    -------
    Currently the OPL CSV don't have an explicit header.
    Here we assume that the header is as follow"

        # key, start, end, obs_key, instrument

    The initial ``key`` is composed of the following parameters:

        <instrument>_<observation_type>_<type>

    """
    return [
        '_'.join(
            [
                obs['INSTRUMENT'],
                obs['OBSERVATION_TYPE'],
                obs['TYPE'],
            ]
        ),
        fmt_datetime(obs.start),
        fmt_datetime(obs.stop),
        obs['OBS_KEY'],
        obs['INSTRUMENT'],
    ]


def save_opl_json(
    fout: str | Path | None,
    *events: EventsList | EventsDict,
    overlap: bool = False,
) -> Path:
    """Export OPL file in JSON format.

    Note
    ----
    Re-use ITL JSON export with OPL schema instead.

    """
    return save_itl_json(fout, *events, overlap=overlap, fmt='OPL')



"""Majis ITL in JSON format sub-module."""

import json
from pathlib import Path

from planetary_coverage.events import EventWindow

from ...schema import MAJIS_SCHEMAS
from .mapping import JSON_MAPPING


def read_itl_json(
    fname: str | Path,
    fmt: str = 'ITL',
    only: str | list | bool | None = 'MAJIS',
) -> list[EventWindow]:
    """Read ITL file in JSON format.

    ITL are validated against Juice SOC and Majis schema.

    By default only the MAJIS events will be reported.
    This can be changed by providing an other instrument filter key,
    a list of keys or an explicit 'ALL', False or None value.

    """
    content = json.loads(Path(fname).read_text())

    for schema in MAJIS_SCHEMAS[fmt]:
        schema.validate(content)

    # Check filename match the one in the file
    filename = Path(content['header']['filename'])
    if not filename or filename.name != Path(fname).name:
        raise FileNotFoundError(filename)

    # Set default values
    defaults = {
        fmt: Path(filename),
    }

    return [
        EventWindow(
            obs['name'],
            parse_json(obs, defaults=defaults),
        )
        for obs in content['timeline']
        # Filter observations by instrument name(s)
        if not only
        or obs['instrument'] in only
        or obs['instrument'] == only
        or only == 'ALL'
    ]


def parse_json(obs: dict, defaults: dict | None = None) -> dict:
    """Map ITL JSON timeline keys to EPS event dictionary."""
    # Extract parameters if present
    params = obs.pop('parameters') if 'parameters' in obs else {}

    # Convert JSON keys to EPS keys
    content = {
        JSON_MAPPING.get(key, key.upper()): value for key, value in (obs | params).items()
    }

    # Add extra keys
    if 'OBSERVATION_TYPE' in content:
        content['PRIME'] = content['OBSERVATION_TYPE'] == 'PRIME'

    # Add default keys if not already present
    if defaults:
        for key, value in defaults.items():
            if key.upper() not in content:
                content[key.upper()] = value

    # Put comments last if present
    content['COMMENTS'] = content.pop('COMMENTS') if 'COMMENTS' in content else ''

    return content



"""MAJIS miscellaneous submodule."""

from .depreciation import depreciated
from .evf import read_evf
from .time import fmt_datetime, get_datetime

__all__ = [
    'read_evf',
    'get_datetime',
    'fmt_datetime',
    'depreciated',
]
